"""Command line arguments shared variables across all modules."""

configfile = None
stdout = None
stderr = None
date = None
tel_id = None
nightsummary = True
reason = None
directory = None
log_directory = None
simulate = None
noninteractive = None
verbose = None
warning = None
nocheck = None
no_dl2 = None
prod_id = None
dl1_prod_id = None
dl2_prod_id = None
append = None
provenance = True
force = None
filter = None
quit = None
test = False
no_calib = False
no_submit = False
seqtoclose = None
run = None
filters = None
